@@ -22,6 +22,7 @@ import (
 	"crypto/x509"
 	"encoding/base64"
 	"encoding/hex"
+	"encoding/pem"
 	"errors"
 	"fmt"
 	"strings"
@@ -31,6 +32,8 @@ import (
 	"github.com/go-openapi/swag"
 
 	"github.com/sigstore/cosign/v2/pkg/cosign"
+	cms "github.com/sigstore/gitsign/internal/fork/ietf-cms"
+	rekoroid "github.com/sigstore/gitsign/internal/rekor/oid"
 	rekor "github.com/sigstore/rekor/pkg/client"
 	"github.com/sigstore/rekor/pkg/generated/client"
 	"github.com/sigstore/rekor/pkg/generated/client/index"
@@ -43,14 +46,16 @@ import (
 	"github.com/sigstore/sigstore/pkg/tuf"
 )
 
-// Verifier represents a mechanism to get and verify Rekor entries for the given Git commit.
+// Verifier represents a mechanism to get and verify Rekor entries for the given Git data.
 type Verifier interface {
 	Verify(ctx context.Context, commitSHA string, cert *x509.Certificate) (*models.LogEntryAnon, error)
+	VerifyInclusion(ctx context.Context, sig []byte, cert *x509.Certificate) (*models.LogEntryAnon, error)
 }
 
 // Writer represents a mechanism to write content to Rekor.
 type Writer interface {
 	Write(ctx context.Context, commitSHA string, sig []byte, cert *x509.Certificate) (*models.LogEntryAnon, error)
+	WriteMessage(ctx context.Context, message, signature []byte, cert *x509.Certificate) (*models.LogEntryAnon, error)
 }
 
 // Client implements a basic rekor implementation for writing and verifying Rekor data.
@@ -74,16 +79,21 @@ func New(url string, opts ...rekor.Option) (*Client, error) {
 	}, nil
 }
 
+// Deprecated: Use [WriteMessage] instead.
 func (c *Client) Write(ctx context.Context, commitSHA string, sig []byte, cert *x509.Certificate) (*models.LogEntryAnon, error) {
+	return c.WriteMessage(ctx, []byte(commitSHA), sig, cert)
+}
+
+func (c *Client) WriteMessage(ctx context.Context, message, signature []byte, cert *x509.Certificate) (*models.LogEntryAnon, error) {
 	pem, err := cryptoutils.MarshalCertificateToPEM(cert)
 	if err != nil {
 		return nil, err
 	}
 	checkSum := sha256.New()
-	if _, err := checkSum.Write([]byte(commitSHA)); err != nil {
+	if _, err := checkSum.Write(message); err != nil {
 		return nil, err
 	}
-	return cosign.TLogUpload(ctx, c.Rekor, sig, checkSum, pem)
+	return cosign.TLogUpload(ctx, c.Rekor, signature, checkSum, pem)
 }
 
 func (c *Client) get(ctx context.Context, data []byte, cert *x509.Certificate) (*models.LogEntryAnon, error) {
@@ -163,6 +173,15 @@ func rekorPubsFromClient(rekorClient *client.Rekor) (*cosign.TrustedTransparency
 	return &publicKeys, nil
 }
 
+// Verify verifies a commit using online verification.
+//
+// This is done by:
+// 1. Searching Rekor for an entry matching the commit SHA + cert.
+// 2. Use the same cert to verify the commit content.
+//
+// Note: While not truly deprecated, using offline verification is generally preferred.
+// This function relies on non-GA behavior of Rekor, and remains for backwards
+// compatibility with older signatures.
 func (c *Client) Verify(ctx context.Context, commitSHA string, cert *x509.Certificate) (*models.LogEntryAnon, error) {
 	e, err := c.get(ctx, []byte(commitSHA), cert)
 	if err != nil {
@@ -225,3 +244,56 @@ func extractCerts(e *models.LogEntryAnon) ([]*x509.Certificate, error) {
 func (c *Client) PublicKeys() *cosign.TrustedTransparencyLogPubKeys {
 	return c.publicKeys
 }
+
+// VerifyInclusion verifies a signature's inclusion in Rekor using offline verification.
+// NOTE: This does **not** verify the correctness of the signature against the content.
+// Prefer using [git.Verify] instead for complete verification.
+func (c *Client) VerifyInclusion(ctx context.Context, sig []byte, cert *x509.Certificate) (*models.LogEntryAnon, error) {
+	// Try decoding as PEM
+	var der []byte
+	if blk, _ := pem.Decode(sig); blk != nil {
+		der = blk.Bytes
+	} else {
+		der = sig
+	}
+	// Parse signature
+	sd, err := cms.ParseSignedData(der)
+	if err != nil {
+		return nil, fmt.Errorf("failed to parse signature: %w", err)
+	}
+
+	// Extract signer from signature. The spec allows for multiple signers, but in
+	// practice for Git commits there's typically only ever 1 signer (the committer).
+	r := sd.Raw()
+	if len(r.SignerInfos) == 0 {
+		return nil, fmt.Errorf("no signers found in signature")
+	}
+	si := r.SignerInfos[0]
+
+	// Double check cert matches the signer. This technically isn't needed
+	// since if this didn't match then the verify below should also fail,
+	// but this helps us distinguish the error in the unlikely event this does happen.
+	if _, err := si.FindCertificate([]*x509.Certificate{cert}); err != nil {
+		return nil, fmt.Errorf("signer certificate mismatch: %w", err)
+	}
+
+	// Get HashedRekord body from the signature.
+	// We are assuming here that the signature has already been authenticated against the
+	// cert, so it is okay to rely the precomputed checksum in the SignerInfo.
+	message, err := si.GetMessageDigestAttribute()
+	if err != nil {
+		return nil, err
+	}
+
+	// Reassemble the tlog entry from the signature pieces.
+	tlog, err := rekoroid.ToLogEntry(ctx, message, si.Signature, cert, si.UnsignedAttrs)
+	if err != nil {
+		return nil, err
+	}
+
+	if err := cosign.VerifyTLogEntryOffline(ctx, tlog, c.PublicKeys()); err != nil {
+		return nil, err
+	}
+
+	return tlog, nil
+}
